// Copyright (C) 2004-2007 Tord Romstad

// This file is part of Glaurung.
//
// Glaurung is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//  
// Glaurung is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Glaurung; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


////
//// Includes
////

#include "movepick.h"
#include "san.h"

#include <cassert>
#include <cstring>

////
//// Local definitions
////

namespace {

  /// Types

  enum Ambiguity {
    AMBIGUITY_NONE,
    AMBIGUITY_FILE,
    AMBIGUITY_RANK,
    AMBIGUITY_BOTH
  };


  /// Functions
  
  Ambiguity move_ambiguity(Position& pos, Move m);
}


////
//// Functions
////


const std::string move_to_san(Position& pos, Move m) {
  std::string str;

  assert(pos.is_ok());
  assert(move_is_ok(m));

  if(m == MOVE_NONE) {
    str = "(none)";
    return str;
  }
  else if(m == MOVE_NULL) {
    str = "(null)";
    return str;
  }
  else if(move_is_long_castle(m))
    str = "O-O-O";
  else if(move_is_short_castle(m))
    str = "O-O";
  else {
    Square from, to;
    Piece pc;

    from = move_from(m);
    to = move_to(m);
    pc = pos.piece_at(move_from(m));

    str = "";

    if(type_of_piece(pc) == PAWN) {
      if(pos.move_is_capture(m))
	str += file_to_char(square_file(move_from(m)));
    }
    else {
      str += piece_type_to_char(type_of_piece(pc), true);

      Ambiguity amb = move_ambiguity(pos, m);
      switch(amb) {

      case AMBIGUITY_NONE:
	break;

      case AMBIGUITY_FILE:
	str += file_to_char(square_file(from));
	break;

      case AMBIGUITY_RANK:
	str += rank_to_char(square_rank(from));
	break;

      case AMBIGUITY_BOTH:
	str += square_to_string(from);
	break;

      default:
	assert(false);
      }
    }

    if(pos.move_is_capture(m))
      str += "x";

    str += square_to_string(move_to(m));

    if(move_promotion(m)) {
      str += "=";
      str += piece_type_to_char(move_promotion(m), true);
    }
  }

  // Is the move check?  We don't use pos.move_is_check(m) here, because
  // Position::move_is_check doesn't detect all checks (not castling moves,
  // promotions and en passant captures).
  UndoInfo u;
  pos.do_move(m, u);
  if(pos.is_check())
    str += pos.is_mate()? "#" : "+";
  pos.undo_move(m, u);

  return str;
}


Move move_from_san(Position& pos, const std::string& movestr) {
  assert(pos.is_ok());
  
  MovePicker mp = MovePicker(pos, false, MOVE_NONE, MOVE_NONE, MOVE_NONE,
			     MOVE_NONE, OnePly);

  // Castling moves
  if(movestr == "O-O-O") {
    Move m;
    while((m = mp.get_next_move()) != MOVE_NONE)
      if(move_is_long_castle(m) && pos.move_is_legal(m))
	return m;
    return MOVE_NONE;
  }
  else if(movestr == "O-O") {
    Move m;
    while((m = mp.get_next_move()) != MOVE_NONE)
      if(move_is_short_castle(m) && pos.move_is_legal(m))
	return m;
    return MOVE_NONE;
  }

  // Normal moves
  const char* cstr = movestr.c_str();
  const char* c;
  char *cc, str[10];
  int i;

  // Initialize str[] by making a copy of movestr with the characters
  // 'x', '=', '+' and '#' removed.  
  cc = str;
  for(i=0, c=cstr; i<10 && *c!='\0' && *c!='\n' && *c!=' '; i++, c++)
    if(!strchr("x=+#", *c)) {
      *cc = strchr("nrq", *c)? toupper(*c) : *c;
      cc++;
    }
  *cc = '\0';

  int left = 0, right = strlen(str) - 1;
  PieceType pt = NO_PIECE_TYPE, promotion;
  Square to;
  File fromFile = FILE_NONE;
  Rank fromRank = RANK_NONE;

  // Promotion?
  if(strchr("BNRQ", str[right])) {
    promotion = piece_type_from_char(str[right]);
    right--;
  }
  else
    promotion = NO_PIECE_TYPE;

  // Find the moving piece:
  if(left < right) {
    if(strchr("BNRQK", str[left])) {
      pt = piece_type_from_char(str[left]);
      left++;
    }
    else
      pt = PAWN;
  }

  // Find the to square:
  if(left < right) {
    if(str[right] < '1' || str[right] > '8' ||
       str[right-1] < 'a' || str[right-1] > 'h')
      return MOVE_NONE;
    to = make_square(file_from_char(str[right-1]), rank_from_char(str[right]));
    right -= 2;
  }
  else
    return MOVE_NONE;

  // Find the file and/or rank of the from square:
  if(left < right) {
    if(strchr("abcdefgh", str[left])) {
      fromFile = file_from_char(str[left]);
      left++;
    }
    if(strchr("12345678", str[left]))
      fromRank = rank_from_char(str[left]);
  }

  // Look for a matching move:
  Move m, move = MOVE_NONE;
  int matches = 0;
  
  while((m = mp.get_next_move()) != MOVE_NONE) {
    bool match = true;
    if(pos.type_of_piece_at(move_from(m)) != pt)
      match = false;
    else if(move_to(m) != to)
      match = false;
    else if(move_promotion(m) != promotion)
      match = false;
    else if(fromFile != FILE_NONE && fromFile != square_file(move_from(m)))
      match = false;
    else if(fromRank != RANK_NONE && fromRank != square_rank(move_from(m)))
      match = false;
    if(match) {
      move = m;
      matches++;
    }
  }
  
  if(matches == 1)
    return move;
  else
    return MOVE_NONE;
}


namespace {

  Ambiguity move_ambiguity(Position& pos, Move m) {
    Square from, to;
    Piece pc;

    from = move_from(m);
    to = move_to(m);
    pc = pos.piece_at(from);

    // King moves are never ambiguous, because there is never two kings of
    // the same color.
    if(type_of_piece(pc) == KING)
      return AMBIGUITY_NONE;

    MovePicker mp = MovePicker(pos, false, MOVE_NONE, MOVE_NONE, MOVE_NONE,
			       MOVE_NONE, OnePly);
    Move mv, moveList[8];
    int i, j, n;

    n = 0;
    while((mv = mp.get_next_move()) != MOVE_NONE)
      if(move_to(mv) == move_to(m) && pos.piece_at(move_from(mv)) == pc
	 && pos.move_is_legal(mv))
	moveList[n++] = mv;
    if(n == 1)
      return AMBIGUITY_NONE;

    j = 0;
    for(i = 0; i < n; i++)
      if(square_file(move_from(moveList[i])) == square_file(from))
	j++;
    if(j == 1)
      return AMBIGUITY_FILE;

    j = 0;
    for(i = 0; i < n; i++)
      if(square_rank(move_from(moveList[i])) == square_rank(from))
	j++;
    if(j == 1)
      return AMBIGUITY_RANK;

    return AMBIGUITY_BOTH;
  }

}
